﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;

namespace MyEzBuilder
{
    class Program
    {
        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        static extern bool CreateHardLink(string lpFileName, string lpExistingFileName, IntPtr lpSecurityAttributes);

        static DateTime RetrieveLinkerTimestamp()
        {
            string filePath = System.Reflection.Assembly.GetCallingAssembly().Location;
            const int c_PeHeaderOffset = 60;
            const int c_LinkerTimestampOffset = 8;
            byte[] b = new byte[2048];
            System.IO.Stream s = null;

            try
            {
                s = new System.IO.FileStream(filePath, System.IO.FileMode.Open, System.IO.FileAccess.Read);
                s.Read(b, 0, 2048);
            }
            finally
            {
                if (s != null)
                {
                    s.Close();
                }
            }

            int i = System.BitConverter.ToInt32(b, c_PeHeaderOffset);
            int secondsSince1970 = System.BitConverter.ToInt32(b, i + c_LinkerTimestampOffset);
            DateTime dt = new DateTime(1970, 1, 1, 0, 0, 0);
            dt = dt.AddSeconds(secondsSince1970);
            dt = dt.AddHours(TimeZone.CurrentTimeZone.GetUtcOffset(dt).Hours);
            return dt;
        }

        [MTAThread]
        static void Main(string[] args)
        {
            string sChooseDir = string.Empty;
            string sAvsInsert = string.Empty;
            string sFile = string.Empty;

            // tool\ini\setup.ini
            bool bLoadAVS = false;
            bool bTextSubDirect = false;
            bool bNormalize = false;
            string sLanguage = "cp950";
            int nCPU_cores = 1;
            bool bUTF8Comment = false;
            bool bChangeCodePage = false;

            // parameter
            bool bFullScreen = false;
            string sChangeInputDir = string.Empty;
            string sChangeOutputDir = string.Empty;
            string sProfileDir = string.Empty;

            DateTime dt = RetrieveLinkerTimestamp();
            string sYear = dt.ToString("yyyy");
            if (!sYear.Equals("2013"))
                sYear = "2013-" + sYear;
            string sMyVer = "MyEzBuilder (" + dt.ToString("yyyy/MM/dd HH:mm:ss") + "), Copyright (c) " + sYear + " Tommy Wu";

            Console.WriteLine("");
            Console.WriteLine(sMyVer);
            Console.WriteLine("");

            // 載入設定檔
            try
            {
                sFile = "tool\\ini\\setup.ini";
                if (File.Exists(sFile))
                {
                    string[] readText = File.ReadAllLines(sFile, Encoding.Default);

                    foreach (string sLine in readText)
                    {
                        try
                        {
                            string[] s = sLine.Split(new Char[] { '=' }, 2);

                            string sItem = s[0].Trim();
                            string sValue = s[1].Trim();

                            if (sItem.Equals("load_avs"))
                                bLoadAVS = !sValue.Equals("0");
                            if (sItem.Equals("textsub_direct"))
                                bTextSubDirect = !sValue.Equals("0");
                            if (sItem.Equals("normalize"))
                                bNormalize = !sValue.Equals("0");
                            if (sItem.Equals("language"))
                                sLanguage = sValue;
                            if (sItem.Equals("CPU_cores"))
                            {
                                try
                                {
                                    nCPU_cores = Convert.ToInt32(sValue);
                                }
                                catch (Exception)
                                {
                                    nCPU_cores = 1;
                                }
                            }
                            if (sItem.Equals("utf8_comment"))
                                bUTF8Comment = !sValue.Equals("0");
                            if (sItem.Equals("change_codepage"))
                                bChangeCodePage = !sValue.Equals("0");
                        }
                        catch (Exception)
                        {
                        }
                    }
                }
            }
            catch (Exception)
            {
            }

            // 載入 tool\avs_insert.txt
            if (bLoadAVS)
            {
                try
                {
                    sFile = "tool\\avs_insert.txt";
                    if (File.Exists(sFile))
                        sAvsInsert = File.ReadAllText(sFile);
                }
                catch (Exception)
                {
                }
            }

            // 讀取傳入的參數
            if (args.GetLength(0) > 0)
            {
                try
                {
                    int max_args = args.GetLength(0);
                    int i;

                    // 更改檔名 
                    if (args[0].Equals("-rename", StringComparison.OrdinalIgnoreCase)) 
                    {
                        try
                        {
                            if (max_args != 3)
                                Console.WriteLine("-rename 參數數量錯誤");
                            else
                                File.Move(args[1], args[2]);
                        }
                        catch (Exception)
                        {
                        }
                        Environment.Exit(0);
                    }
                    for (i = 0; i < max_args; i++)
                    {
                        if (args[i].Equals("-fullscreen", StringComparison.OrdinalIgnoreCase))
                        {
                            bFullScreen = true;
                            Console.WriteLine("fullscreen => true");
                        }
                        else if (args[i].Equals("-profile", StringComparison.OrdinalIgnoreCase))
                        {
                            i++;
                            if (i < max_args)
                            {
                                sProfileDir = args[i];
                                if (!sProfileDir.Substring(sProfileDir.Length).Equals("\\"))
                                    sProfileDir = sProfileDir + "\\";
                                Console.WriteLine("profile => " + sProfileDir);
                            }
                        }
                        else if (args[i].Equals("-change_input_dir", StringComparison.OrdinalIgnoreCase))
                        {
                            i++;
                            if (i < max_args)
                            {
                                sChangeInputDir = args[i];
                                if (!sChangeInputDir.Substring(sChangeInputDir.Length).Equals("\\"))
                                    sChangeInputDir = sChangeInputDir + "\\";
                                Console.WriteLine("change_input_dir => " + sChangeInputDir);
                            }
                        }
                        else if (args[i].Equals("-change_output_dir", StringComparison.OrdinalIgnoreCase))
                        {
                            i++;
                            if (i < max_args)
                            {
                                sChangeOutputDir = args[i];
                                if (!sChangeOutputDir.Substring(sChangeOutputDir.Length).Equals("\\"))
                                    sChangeOutputDir = sChangeOutputDir + "\\";
                                Console.WriteLine("change_output_dir => " + sChangeOutputDir);
                            }
                        }
                    }
                }
                catch (Exception)
                {
                }
            }

            if (string.IsNullOrEmpty(sProfileDir))
            {
                // 未指定 profile
                string sAccessDir = string.Empty;
                bool bDone = false;

                while (!bDone)
                {
                    try
                    {
                        Console.WriteLine("請選擇您要使用的 profile");
                        Console.WriteLine("");

                        if (sAccessDir != string.Empty)
                        {
                            Console.WriteLine("[.] \" 回主選單 \"");
                        }

                        List<string> profileItems = new List<string>();
                        int nProfileCnt = 0;
                        string sProfileRoot = "tool\\profile\\" + sAccessDir;
                        string[] dirNames = Directory.GetDirectories(sProfileRoot, "*", SearchOption.TopDirectoryOnly);
                        foreach (string sDir in dirNames)
                        {
                            string sSubDir = sDir.Substring(sProfileRoot.Length);
                            string sComment = sDir;
                            sFile = sDir + "\\comment.txt";
                            if (File.Exists(sFile))
                            {
                                string[] readText;
                                if (bUTF8Comment)
                                    readText = File.ReadAllLines(sFile, Encoding.UTF8);
                                else
                                    readText = File.ReadAllLines(sFile, Encoding.Default);
                                sComment = readText[0].Trim();
                            }
                            nProfileCnt++;
                            sFile = sDir + "\\dir_tag.txt";
                            if (File.Exists(sFile))
                            {
                                Console.WriteLine("[" + nProfileCnt.ToString() + "][DIR] " + sComment);
                            }
                            else
                            {
                                Console.WriteLine("[" + nProfileCnt.ToString() + "] " + sComment);
                            }
                            profileItems.Add(sSubDir);
                        }

                        Console.WriteLine("");

                        if (nProfileCnt == 0)
                        {
                            Console.WriteLine("沒有任何 profile 在此目錄");
                            Console.WriteLine("離開...");
                            Environment.Exit(0);
                        }

                        Console.Write("選輸入你的選擇 : ");
                        string sChoose = Console.ReadLine();
                        if (sChoose.Equals("."))
                        {
                            sAccessDir = string.Empty;
                            Console.WriteLine("");
                            continue;
                        }
                        int nChoose = 0;
                        try
                        {
                            nChoose = Convert.ToInt32(sChoose);
                        }
                        catch (Exception)
                        {
                            nChoose = 0;
                        }
                        if (nChoose < 1 || nChoose > nProfileCnt)
                        {
                            Console.WriteLine("");
                            continue;
                        }

                        sAccessDir = sAccessDir + profileItems[nChoose - 1] + "\\";
                        if (File.Exists(sProfileRoot + sAccessDir + "dir_tag.txt"))
                        {
                            Console.WriteLine("");
                            continue;
                        }
                        sChooseDir = sProfileRoot + sAccessDir;
                        bDone = true;
                    }
                    catch (Exception)
                    {
                    }
                }
            }
            else
            {
                // 指定使用某個 profile
                try
                {
                    string sComment = sProfileDir;
                    sFile = "tool\\profile\\" + sProfileDir + "comment.txt";
                    if (File.Exists(sFile))
                    {
                        string[] readText;
                        if (bUTF8Comment)
                            readText = File.ReadAllLines(sFile, Encoding.UTF8);
                        else
                            readText = File.ReadAllLines(sFile, Encoding.Default);
                        sComment = readText[0].Trim();
                    }

                    sChooseDir = "tool\\profile\\" + sProfileDir;
                }
                catch (Exception)
                {
                }
            }

            string sBatWork = string.Empty;
            string sAvsDeal = string.Empty;
            string sFontStyle = string.Empty;
            string[] profileSetup = null;

            // 批次指令內容
            sFile = sChooseDir + "bat_work.txt";
            if (!File.Exists(sFile))
            {
                Console.WriteLine("找不到 " + sFile);
                Console.WriteLine("離開...");
                Environment.Exit(0);
            }
            sBatWork = File.ReadAllText(sFile);

            // profile 參數設定
            sFile = sChooseDir + "profile_setup.txt";
            if (File.Exists(sFile))
                profileSetup = File.ReadAllLines(sFile, Encoding.Default);

            // 字型 style
            sFile = sChooseDir + "font_style.txt";
            if (File.Exists(sFile))
                sFontStyle = File.ReadAllText(sFile);

            // avisynth 設定檔
            sFile = sChooseDir + "avs_deal.txt";
            if (File.Exists(sFile))
                sAvsDeal = File.ReadAllText(sFile);

            string sAvsDealOrig = sAvsDeal;
            string sBatWorkOrig = sBatWork;

            // 轉換 profile 的參數
            if (profileSetup != null)
            {
                foreach (string sLine in profileSetup)
                {
                    try
                    {
                        string[] s = sLine.Split(new Char[] { '=' }, 2);

                        sAvsDealOrig = sAvsDealOrig.Replace(s[0].Trim(), s[1].Trim());
                        sBatWorkOrig = sBatWorkOrig.Replace(s[0].Trim(), s[1].Trim());
                    }
                    catch (Exception)
                    {
                    }
                }
            }

            // 輸入的檔案所在目錄
            string sMediaFileDir = "MediaFile\\";
            if (!string.IsNullOrEmpty(sChangeInputDir))
                sMediaFileDir = sChangeInputDir;
            // 輸出的檔案所在目錄
            string sDoneFileDir = "DoneFile\\";
            if (!string.IsNullOrEmpty(sChangeOutputDir))
                sDoneFileDir = sChangeOutputDir;

            // 產生的批次檔內容
            string sBuildBatData = string.Empty;

            // 切換到 utf-8 (65001) codepage
            int nCodePage = Console.OutputEncoding.CodePage;
            if (bChangeCodePage)
                sBuildBatData = "chcp 65001\n";

            string[] fileNames = Directory.GetFiles(sMediaFileDir, "*", SearchOption.TopDirectoryOnly);
            int nCount = 0;
            foreach (string sMediaFile in fileNames)
            {
                string sExt = Path.GetExtension(sMediaFile);

                // 字幕檔不處理
                if (sExt.Equals(".srt", StringComparison.OrdinalIgnoreCase) ||
                    sExt.Equals(".style", StringComparison.OrdinalIgnoreCase) ||
                    sExt.Equals(".db", StringComparison.OrdinalIgnoreCase) ||
                    sExt.Equals(".idx", StringComparison.OrdinalIgnoreCase) ||
                    sExt.Equals(".sub", StringComparison.OrdinalIgnoreCase) ||
                    sExt.Equals(".ass", StringComparison.OrdinalIgnoreCase) ||
                    sExt.Equals(".ssa", StringComparison.OrdinalIgnoreCase))
                    continue;

                // 檔案名稱
                string sFileName = Path.GetFileName(sMediaFile);
                // 未含 extension 的檔案名稱
                string sMainName = Path.GetFileNameWithoutExtension(sMediaFile);

                // 轉換 avisyntch 設定檔的內容
                sAvsDeal = sAvsDealOrig;
                sAvsDeal = sAvsDeal.Replace("AvsNumberLoad", nCount.ToString());
                sAvsDeal = sAvsDeal.Replace("File_To_Load", sFileName);
                sAvsDeal = sAvsDeal.Replace("MainFileLoad", sMainName);
                sAvsDeal = sAvsDeal.Replace("ExtFileLoad", sExt);
                sAvsDeal = sAvsDeal.Replace("MediaFileDir", sMediaFileDir);
                sAvsDeal = sAvsDeal.Replace("DoneFileDir", sDoneFileDir);
                if (bFullScreen)
                    sAvsDeal = sAvsDeal.Replace("FL_SETUP", "true");
                else
                    sAvsDeal = sAvsDeal.Replace("FL_SETUP", "false");

                // 轉換批次檔的內容
                sBatWork = sBatWorkOrig;
                sBatWork = sBatWork.Replace("AvsNumberLoad", nCount.ToString());
                sBatWork = sBatWork.Replace("File_To_Load", sFileName);
                sBatWork = sBatWork.Replace("MainFileLoad", sMainName);
                sBatWork = sBatWork.Replace("ExtFileLoad", sExt);
                sBatWork = sBatWork.Replace("MediaFileDir", sMediaFileDir);
                sBatWork = sBatWork.Replace("DoneFileDir", sDoneFileDir);
                sBatWork = sBatWork.Replace("CPU_CORES", nCPU_cores.ToString());

                sBuildBatData = sBuildBatData + sBatWork + "\n";

                string sTempFile;
                string sName;
                string sDel = string.Empty;

                // 產生檔案 hardlink 
                sTempFile = "temp\\temp_media_" + nCount.ToString() + sExt;
                sName = sMediaFileDir + sFileName;
                CreateHardLink(sTempFile, sName, IntPtr.Zero);
                sDel = sDel + "rem delete files\ndel /q \"" + sTempFile + "\"\n";

                bool bLoaded = false;

                // 處理 idx/sub 的字幕
                sName = sMediaFileDir + sMainName + ".idx";
                if (File.Exists(sName)) 
                {
                    sAvsDeal = sAvsDeal + "\nvobsub(\"temp_media_" + nCount.ToString() + ".idx\")\n";

                    sTempFile = "temp\\temp_media_" + nCount.ToString() + ".idx";
                    CreateHardLink(sTempFile, sName, IntPtr.Zero);
                    sDel = sDel + "del /q \"" + sTempFile + "\"\n";

                    sTempFile = "temp\\temp_media_" + nCount.ToString() + ".sub";
                    sName = sMediaFileDir + sMainName + ".sub";
                    CreateHardLink(sTempFile, sName, IntPtr.Zero);
                    sDel = sDel + "del /q \"" + sTempFile + "\"\n";

                    bLoaded = true;
                }

                // 處理 ass 的字幕
                sName = sMediaFileDir + sMainName + ".ass";
                if (!bLoaded && File.Exists(sName)) 
                {
                    sAvsDeal = sAvsDeal + "\ntextsub(\"temp_media_" + nCount.ToString() + ".ass\")\n";

                    sTempFile = "temp\\temp_media_" + nCount.ToString() + ".ass";
                    CreateHardLink(sTempFile, sName, IntPtr.Zero);
                    sDel = sDel + "del /q \"" + sTempFile + "\"\n";

                    bLoaded = true;
                }

                // 處理 ssa 的字幕
                sName = sMediaFileDir + sMainName + ".ssa";
                if (!bLoaded && File.Exists(sName)) 
                {
                    sAvsDeal = sAvsDeal + "\ntextsub(\"temp_media_" + nCount.ToString() + ".ssa\")\n";

                    sTempFile = "temp\\temp_media_" + nCount.ToString() + ".ssa";
                    CreateHardLink(sTempFile, sName, IntPtr.Zero);
                    sDel = sDel + "del /q \"" + sTempFile + "\"\n";

                    bLoaded = true;
                }

                // 處理 srt 的字幕
                sName = sMediaFileDir + sMainName + ".srt";
                if (!bLoaded && File.Exists(sName)) 
                {
                    sAvsDeal = sAvsDeal + "\ntextsub(\"temp_media_" + nCount.ToString() + ".srt\")\n";

                    sTempFile = "temp\\temp_media_" + nCount.ToString() + ".srt";
                    CreateHardLink(sTempFile, sName, IntPtr.Zero);
                    sDel = sDel + "del /q \"" + sTempFile + "\"\n";

                    File.WriteAllText(sTempFile + ".style", sFontStyle);
                    sDel = sDel + "del /q \"" + sTempFile + ".style\"\n";

                    bLoaded = true;
                }

                // 對聲音正規化處理 (如果不是透過 avisynth 處理就沒效果)
                if (bNormalize) 
                {
                    sAvsDeal = sAvsDeal + "\nnormalize()\n";
                }
                sAvsDeal = sAvsDeal + "\nreturn last\n";

                // 附加刪除檔案的指令
                sBuildBatData = sBuildBatData + "\n" + sDel;

                // 產生 avisynth 設定檔
                File.WriteAllText("temp\\temp_avs_" + nCount.ToString() + ".avs", sAvsInsert + sAvsDeal);
                nCount++;
            }

            // 切換為原本的 codepage
            if (bChangeCodePage)
                sBuildBatData = sBuildBatData + "\nchcp " + nCodePage.ToString() + "\n";

            if (nCount > 0)
            {
                // 產生批次檔
                File.WriteAllText("tool\\build.bat", sBuildBatData);
                // 執行批次檔
                System.Diagnostics.Process.Start("tool\\build.bat");
            }
        }
    }
}
